////////////////////////////////////////////////////////
// filename: CList.cpp
// author: Chafumi Touji
// version: 1.0.1
// date: 2012/04/09
////////////////////////////////////////////////////////

#include "ChaSound.h"
#include "CListB.h"

////////////////////////////////////////////////////////////////////////////////////////

CListB::CListB(){

	CurrentNum = 0;
	ListSize = 0;
	Head = Tail = CurrentPos = NULL;
}

CListB::~CListB(){

	DeleteAll();
	CurrentNum = ListSize = 0;
	Head = Tail = CurrentPos = NULL;
}

/////////////////////////////////////////////////////////////////////////////////////////

CListDataB *CListB::GetData(){
	
	return CurrentPos;
}

CListDataB *CListB::GetHead(){

	return Head;
}

CListDataB *CListB::GetTail(){

	return Tail;
}

CListDataB *CListB::SetNext(){

	if( ListSize <= 0 )
		return NULL;

	if( CurrentPos == NULL )
		return NULL;

	CurrentPos = CurrentPos->Next;
	CurrentNum++;

	return CurrentPos;
}

CListDataB *CListB::SetPrev(){

	if( ListSize <= 0 )
		return NULL;

	if( CurrentPos == NULL )
		return NULL;

	CurrentPos = CurrentPos->Prev;
	CurrentNum--;

	return CurrentPos;
}

CListDataB *CListB::InsData( void *data ){

	return InsData(data, 0);
}

CListDataB *CListB::InsTail( void *data ){

	return InsTail(data, 0);
}

CListDataB *CListB::InsHead( void *data ){

	return InsHead(data, 0);
}

CListDataB *CListB::InsData(void *data, int id ){

	if (CurrentPos == Tail)
		return InsTail(data, id );

	CListDataB *new_element = NULL;
	new_element = new CListDataB();
	new_element->Data = data;
	new_element->ID = id;
	if (CurrentPos == NULL && ListSize <= 0){
		Head = Tail = CurrentPos = new_element;
		CurrentNum = 1;
	}
	else{
		if (CurrentPos == Tail)
			Tail = new_element;
		new_element->Next = CurrentPos->Next;
		new_element->Prev = CurrentPos;
		if (CurrentPos->Next != NULL)
			CurrentPos->Next->Prev = new_element;
		CurrentPos->Next = new_element;
	}

	ListSize++;

	return new_element;
}

CListDataB *CListB::InsTail(void *data, int id ){

	CListDataB *new_element = NULL;
	new_element = new CListDataB();
	new_element->Data = data;
	new_element->ID = id;
	if (CurrentPos == NULL && ListSize <= 0){
		Head = Tail = CurrentPos = new_element;
		CurrentNum = 1;
	}
	else{
		new_element->Next = NULL;
		new_element->Prev = Tail;
		Tail->Next = new_element;
		Tail = new_element;
	}
	ListSize++;

	return new_element;
}

CListDataB *CListB::InsHead(void *data, int id ){

	CListDataB *new_element = NULL;
	new_element = new CListDataB();
	new_element->Data = data;
	new_element->ID = id;
	if (CurrentPos == NULL && ListSize <= 0){
		Head = Tail = CurrentPos = new_element;
		CurrentNum = 1;
	}
	else{
		new_element->Next = Head;
		new_element->Prev = NULL;
		Head->Prev = new_element;
		Head = new_element;
		CurrentNum++;
	}
	ListSize++;

	return new_element;
}

bool CListB::SetAt(int num){

	if( num <= 0 )
		return NULL;

	if( num > ListSize )
		return NULL;

	if( CurrentPos == NULL || num == 1 ){
		CurrentPos = Head;
		CurrentNum = 1;
	}

	if( num > CurrentNum ){
		while( num != CurrentNum )
			SetNext();
	}
	else if( num < CurrentNum ){
		while( num != CurrentNum )
			SetPrev();
	}

	return true;
}

CListDataB *CListB::GetAt( int num ){

	//num͂PListSize܂ŁB

	if( num <= 0 )
		return NULL;

	if( num > ListSize )
		return NULL;

	CListDataB *tmp = NULL;
	tmp = CurrentPos;
	if( num < CurrentNum ){
		while( num != CurrentNum ){
			tmp = tmp->Prev;
			num++;
		}
	}
	else if( num > CurrentNum ){
		while( num != CurrentNum ){
			tmp = tmp->Next;
			num--;
		}
	}

	return tmp;
}

CListDataB *CListB::GetByID(int id){

	for (int i = 0; i <= ListSize - 1; i++){
		CListDataB *tmp = GetAt(i + 1);
		if (tmp->ID == id)
			return tmp;
	}

	return NULL;
}

bool CListB::DeleteAll(){
	
	CurrentPos = Head;
	CListDataB *tmp = NULL;
	tmp = CurrentPos;
	while( tmp != NULL )
		tmp = DeleteData();

	return true;
}

CListDataB *CListB::Delete(int num){

	int new_current = CurrentNum;
	if (new_current > num || new_current == ListSize )
		new_current--;

	SetAt(num);
	DeleteData();
	SetAt(new_current);
	CurrentNum = new_current;

	return NULL;
}

CListDataB *CListB::DeleteByID(int id){

	for (int i = 0; i <= ListSize - 1; i++){
		CListDataB *tmp = GetAt(i + 1);
		if (tmp->ID == id){
			SetAt(i + 1);
			DeleteData();
			if (CurrentNum > i + 1)
				CurrentNum = i + 1;
			if (ListSize <= 0)
				CurrentNum = 0;

			return NULL;
		}
	}

	return NULL;
}

CListDataB *CListB::DeleteData(){

	if( ListSize <= 0 )
		return NULL;

	if( Head == NULL )
		return NULL;

	if( CurrentPos == Head ){
		Head = CurrentPos->Next;
		delete CurrentPos;
		CurrentPos = Head;
		if( Head != NULL )
			Head->Prev = NULL;
		ListSize--;
	}
	else if( CurrentPos == Tail ){
		Tail = CurrentPos->Prev;
		delete CurrentPos->Data;
		delete CurrentPos;
		CurrentPos = Tail;
		Tail->Next = NULL;
		CurrentNum--;
		ListSize--;
	}
	else{
		CListDataB *tmp = NULL;
		tmp = CurrentPos->Prev;
		CurrentPos->Next->Prev = CurrentPos->Prev;
		CurrentPos->Prev->Next = CurrentPos->Next;
		delete CurrentPos->Data;
		delete CurrentPos;
		CurrentPos = tmp;
		CurrentNum--;
		ListSize--;
	}

	return CurrentPos;
}

CListDataB *CListB::DeleteHead(){

	int current_num = CurrentNum;
	SetAt( 1 );
	DeleteData();
	SetAt( current_num - 1 );
	return CurrentPos;
}


bool CListB::IsHead(){

	if( ListSize <= 0 )
		return false;

	if( CurrentPos == Head )
		return true;

	return false;
}

bool CListB::IsTail(){

	if( ListSize <= 0 )
		return false;

	if( CurrentPos == Tail )
		return true;

	return false;
}

bool CListB::Find( void *data ){

	if( ListSize <= 0 )
		return false;

	SetAt( 1 );
	do{
		void *d = GetData()->Data;
		if( d == data )
			return true;
		SetNext();
	}while( IsTail() == false );

	return false;
}

